# Copyright 2019 Canonical Ltd.  This software is licensed under the
# Apache License version 2.0 (see the file LICENSE).

from __future__ import absolute_import, print_function, unicode_literals

from collections import defaultdict
import os.path
import re
import signal
import subprocess

from charmhelpers.contrib.charmsupport.nrpe import NRPE
from charmhelpers.core import (
    hookenv,
    unitdata,
)
from charmhelpers.core.host import (
    service_running,
    service_start,
    service_stop,
)
from charmhelpers.core.templating import render
from charms.apt import queue_install
from charms.reactive import (
    clear_flag,
    hook,
    set_flag,
    when,
    when_any,
    when_not,
)
from charms.reactive.relations import (
    endpoint_from_flag,
    endpoint_from_name,
)


def reload_nfs_ganesha():
    """Reload nfs-ganesha if it is running.

    This should really be done using service_reload, but with nfs-ganesha
    2.6.0 in bionic the systemd unit configuration ends up making a D-Bus
    call that doesn't seem to be hooked up to anything at the other end
    (this is fixed in >= 2.7.0, so disco and newer).
    """
    pid_str = subprocess.check_output(
        ['systemctl', 'show', '--value', '-p', 'MainPID',
         'nfs-ganesha.service'], universal_newlines=True).strip()
    if pid_str:
        pid = int(pid_str)
        if pid:
            os.kill(pid, signal.SIGHUP)


@when_not('nfs-ganesha.requested')
def install_nfs_ganesha():
    # XXX cjwatson 2019-07-15: We should probably add support for other
    # FSALs at some point.
    queue_install(['nfs-ganesha-vfs'])
    set_flag('nfs-ganesha.requested')


@when('apt.installed.nfs-ganesha-vfs')
@when_not('nfs-ganesha.installed')
def mark_installed():
    set_flag('nfs-ganesha.installed')


@when('endpoint.nfs-ganesha.joined')
@when_any('nfs-ganesha.changed',
          'config.changed.storage_root',
          'config.changed.squash')
def configure_nfs_ganesha():
    hookenv.status_set('maintenance', 'Updating NFS-Ganesha configuration')
    clear_flag('nfs-ganesha.ready')

    mount_interface = endpoint_from_flag('endpoint.nfs-ganesha.joined')
    if mount_interface is None:
        hookenv.log('No mount interface')
        return

    config = hookenv.config()
    storage_root = config.get('storage_root')
    squash = config.get('squash')

    mount_addresses = defaultdict(set)

    # Manage a persistent record of which export IDs have been assigned.  At
    # the moment per-unit persistence seems to be OK, since migrating
    # between active nfs-ganesha units requires clients to remount anyway;
    # but if we start doing HA or clustering then we may need to come up
    # with something better.
    export_ids_key = 'nfs-ganesha.exports'
    export_ids = unitdata.kv().get(export_ids_key) or {}
    max_export_id = max(export_ids.values()) if export_ids else 0

    for mount in mount_interface.get_mount_requests():
        name = mount['application_name']
        if not name:
            continue
        path = os.path.join(storage_root, name)
        if not os.path.exists(path):
            hookenv.log('Creating export data path {}'.format(path))
            os.makedirs(path)
            # We can't know what users on clients will need to write to this
            # path, so filesystem security has to be up to clients.
            os.chmod(path, 0o777)
        mount_addresses[name].update(mount['addresses'])
        if name not in export_ids:
            max_export_id += 1
            export_ids[name] = max_export_id

    unitdata.kv().set(export_ids_key, export_ids)

    filename = '/etc/ganesha/ganesha.conf'
    if mount_addresses:
        context = {
            'mounts': [{
                'export_id': export_ids[name],
                'application_name': name,
                'mountpoint': os.path.join(storage_root, name),
                'addresses': sorted(mount_addresses[name]),
                'squash': squash,
            } for name in sorted(mount_addresses)],
        }
        render('ganesha.conf.j2', filename, context)
        hookenv.log('Rendering template to {}'.format(filename))
    else:
        os.remove(filename)

    set_flag('nfs-ganesha.refresh-mounts')


@when('endpoint.nfs-ganesha.joined')
@when_any('nfs-ganesha.refresh-mounts',
          'config.changed.mount_options',
          'config.changed.active_units')
def refresh_mounts():
    hookenv.status_set('maintenance', 'Updating NFS mounts')
    clear_flag('nfs-ganesha.ready')

    mount_interface = endpoint_from_flag('endpoint.nfs-ganesha.joined')
    if mount_interface is None:
        hookenv.log('No mount interface')
        return

    service_is_running = service_running('nfs-ganesha')
    reload_nfs_ganesha()

    config = hookenv.config()
    mount_options = config['mount_options']
    active_ip = None
    active_units = config['active_units']
    need_service = True

    if active_units:
        # Work out which unit should be active so that we can publish its
        # address on the mount relation.  We pick the first item in
        # active_units that exists.
        active_ip = None
        peer_endpoint = endpoint_from_name('peer')
        if peer_endpoint is not None:
            peer_info = peer_endpoint.get_peer_info()
            for active_unit in re.split(r'\s*,\s*', active_units):
                if active_unit in peer_info:
                    active_ip = peer_info[active_unit]['address']
                    hookenv.log(
                        'Active unit found: {} ({})'.format(
                            active_unit, active_ip))
                    break
    else:
        active_ip = hookenv.unit_private_ip()
    need_service = hookenv.unit_private_ip() == active_ip

    # Start nfs-ganesha if this unit is active.
    if need_service and not service_is_running:
        try:
            service_start('nfs-ganesha')
        except subprocess.CalledProcessError as e:
            hookenv.log(e)
            hookenv.log('Unable to start nfs-ganesha.')
            return

    # Publish our details.
    if active_ip is not None:
        # Publish details of the active unit.
        mount_response_common = {
            'hostname': active_ip,
            'fstype': 'nfs',
            'options': mount_options,
        }
    else:
        # There are no active units at all.  Clear out previous responses so
        # that requirers know they need to unmount.
        hookenv.log('No active units')
        mount_response_common = {
            'hostname': None,
            'fstype': None,
            'options': None,
        }
    mount_responses = []
    for mount in mount_interface.get_mount_requests():
        if not mount['application_name']:
            continue
        if active_ip is not None:
            path = '/' + mount['application_name']
        else:
            path = None
        mount_response = {
            'export_name': mount['application_name'],
            'identifier': mount['identifier'],
            'mountpoint': path,
        }
        mount_response.update(mount_response_common)
        mount_responses.append(mount_response)
    mount_interface.configure(mount_responses)

    # Stop nfs-ganesha if this unit is inactive.
    if not need_service and service_is_running:
        service_stop('nfs-ganesha')

    clear_flag('nfs-ganesha.refresh-mounts')


@when('nfs-ganesha.installed')
@when_not('nfs-ganesha.changed',
          'nfs-ganesha.refresh-mounts',
          'nfs-ganesha.ready')
def ready():
    hookenv.status_set('active', 'NFS-Ganesha ready')
    set_flag('nfs-ganesha.ready')


@when('nrpe-external-master.available', 'nfs-ganesha.installed')
@when_not('nfs-ganesha.nrpe.published')
def nrpe_available():
    nrpe_compat = NRPE()
    nrpe_compat.add_check(
        shortname='nfs_port',
        description='NFS server listening check',
        check_cmd='check_tcp -H localhost -p 2049')
    nrpe_compat.write()
    set_flag('nfs-ganesha.nrpe.published')


@hook('nrpe-external-master-relation-changed')
def nrpe_changed(*args):
    # Changed relation data might be nagios_hostname, so republish the
    # check.
    nrpe_compat = NRPE()
    nrpe_compat.remove_check(shortname='nfs_port')
    clear_flag('nfs-ganesha.nrpe.published')


@when('nfs-ganesha.nrpe.published')
@when_not('nrpe-external-master.available')
def nrpe_unavailable():
    clear_flag('nfs-ganesha.nrpe.published')
